import numpy as np
import pytest
import simcardems
from simcardems import save_load_functions as slf


@pytest.fixture
def dummyfile():
    path = "tmp_h5pyfile.h5"
    simcardems.utils.remove_file(path)
    yield path
    simcardems.utils.remove_file(path)


def tests_h5pyfile(dummyfile):

    h5group = "test"
    data1 = [1, 2, 3, 4]
    data2 = [4, 5, 6]

    with slf.h5pyfile(dummyfile, "w") as h5file:
        group = h5file.create_group(h5group)
        group.create_dataset("data1", data=data1)

    with slf.h5pyfile(dummyfile, "a") as h5file:
        h5file[h5group].create_dataset("data2", data=data2)

    with slf.h5pyfile(dummyfile, "r") as h5file:
        assert h5group in h5file

        assert "data1" in h5file[h5group]
        assert np.isclose(h5file[h5group]["data1"][:], data1).all()

        assert "data2" in h5file[h5group]
        assert np.isclose(h5file[h5group]["data2"][:], data2).all()


@pytest.mark.parametrize(
    "data",
    (
        {},
        {"a": 1, "b": 2},
        {"a": 1.0, "b": 2},
        {"a": 1.0, "b": 2, "c": "three"},
        {"a": 1.0, "b": 2, "c": "three", "d": [1, 2]},
    ),
)
def test_dict_to_h5(data, dummyfile):

    h5group = "testgroup"
    slf.dict_to_h5(data, dummyfile, h5group)

    with slf.h5pyfile(dummyfile, "r") as h5file:
        loaded_data = slf.h5_to_dict(h5file[h5group])

    assert loaded_data == data



from pathlib import Path

import dolfin
from simcardems import geometry
from simcardems import slabgeometry

here = Path(__file__).absolute().parent


def test_create_slab_geometry_normal():
    parameters = {"lx": 1, "ly": 1, "lz": 1, "dx": 1, "num_refinements": 1}
    geo = slabgeometry.SlabGeometry(parameters=parameters)

    assert geo.mechanics_mesh.num_cells() == 6
    assert geo.ep_mesh.num_cells() == 48
    for k, v in parameters.items():
        assert geo.parameters[k] == v
    assert geo.num_refinements == 1


def test_create_slab_geometry_with_mechanics_mesh():

    parameters = {"lx": 1, "ly": 1, "lz": 1, "dx": 1, "num_refinements": 1}
    mesh = dolfin.UnitCubeMesh(1, 1, 1)
    geo = slabgeometry.SlabGeometry(
        parameters=parameters,
        mechanics_mesh=mesh,
    )

    assert geo.mechanics_mesh is mesh
    assert geo.ep_mesh.num_cells() == 48
    for k, v in parameters.items():
        assert geo.parameters[k] == v
    assert geo.num_refinements == 1


def test_load_geometry():
    mesh_folder = here / ".." / "demos" / "geometries"
    mesh_path = mesh_folder / "slab.h5"
    schema_path = mesh_folder / "slab.json"
    geo = geometry.load_geometry(mesh_path=mesh_path, schema_path=schema_path)
    assert geo.mesh.num_cells() == 202
    assert geo.ep_mesh.num_cells() == 1616
    assert isinstance(geo, slabgeometry.SlabGeometry)
    assert geo.stimulus_domain.marker == 1


def test_dump_geometry(tmp_path):
    mesh_folder = here / ".." / "demos" / "geometries"
    mesh_path = mesh_folder / "slab.h5"
    schema_path = mesh_folder / "slab.json"
    geo = geometry.load_geometry(mesh_path=mesh_path, schema_path=schema_path)
    outpath = tmp_path / "state.h5"
    geo.dump(outpath)

    dumped_geo = geometry.load_geometry(
        mesh_path=outpath,
        schema_path=outpath.with_suffix(".json"),
    )
    assert dumped_geo == geo


def test_geometry_with_custom_stimulus_domain():
    parameters = {"lx": 1, "ly": 1, "lz": 1, "dx": 0.5, "num_refinements": 1}
    mesh = dolfin.UnitCubeMesh(2, 2, 2)

    def stimulus_domain(mesh):
        marker = 42
        subdomain = dolfin.CompiledSubDomain("x[0] < 0.6")
        domain = dolfin.MeshFunction("size_t", mesh, mesh.topology().dim())
        domain.set_all(0)
        subdomain.mark(domain, marker)
        return geometry.StimulusDomain(domain=domain, marker=marker)

    geo = slabgeometry.SlabGeometry(
        parameters=parameters,
        mechanics_mesh=mesh,
        stimulus_domain=stimulus_domain,
    )
    assert geo.stimulus_domain.marker == 42



import math
import typing


class TimeStepper:
    def __init__(
        self,
        *,
        t0: float,
        T: float,
        dt: float,
        use_ns: bool = True,
        st_progress: typing.Any = None,
    ) -> None:
        """Initialize time stepper

        Parameters
        ----------
        t0 : float
            Start time in milliseconds
        T : float
            End time in milliseconds
        dt : float
            Time step
        use_ns : bool, optional
            Whether to return the time in nanoseconds, by default True
        st_progress:
            Streamlit progress bar
        """

        self._use_ns = use_ns
        self._st_progress = st_progress

        if use_ns:
            self.t0 = TimeStepper.ms2ns(t0)
            self.T = TimeStepper.ms2ns(T)
            self.dt = TimeStepper.ms2ns(dt)
        else:
            self.t0 = t0
            self.T = T
            self.dt = dt

        self.reset()

    def reset(self):
        self.t = self.t0
        self.step = 0

    @property
    def T(self) -> float:
        return self._T

    @T.setter
    def T(self, T: float) -> None:
        if self.t0 > T:
            raise ValueError("Start time has to be lower than end time")
        self._T = T

    @property
    def dt(self) -> float:
        return self._dt

    @dt.setter
    def dt(self, dt: float) -> None:
        dt = min(self.T - self.t0, dt)
        self._dt = dt

    @property
    def total_steps(self) -> int:
        if math.isclose(self.T, self.t0):
            return 0
        return round((self.T - self.t0) / self.dt)

    def __iter__(self):
        if self.T is None:
            raise RuntimeError("Please assign an end time to time stepper")
        while self.t < self.T:

            prev_t = self.t
            self.t = min(self.t + self.dt, self.T)
            self.step += 1
            if self._st_progress is not None:
                self._st_progress.progress(self.step / self.total_steps)
            yield prev_t, self.t

    @staticmethod
    def ns2ms(t: float) -> float:
        """Convert nanoseconds to milliseconds

        Parameters
        ----------
        t : float
            The time in nanoseconds

        Returns
        -------
        float
            Time in milliseconds
        """
        return t * 1e-6

    @staticmethod
    def ms2ns(t: float) -> float:
        """Convert from milliseconds to nanoseconds

        Parameters
        ----------
        t : float
            Time in milliseconds

        Returns
        -------
        float
            Time in nanoseconds
        """
        return int(t * 1e6)



from . import active_model
from . import cell_model
from . import em_model
from .active_model import LandModel as ActiveModel
from .cell_model import ORdmmLandFull as CellModel
from .em_model import EMCoupling

loggers = [
    "simcardems.models.fully_coupled_ORdmm_Land.cell_model.logger",
    "simcardems.models.fully_coupled_ORdmm_Land.em_model.logger",
    "simcardems.models.fully_coupled_ORdmm_Land.active_model.logger",
]

__all__ = [
    "EMCoupling",
    "CellModel",
    "ActiveModel",
    "loggers",
    "em_model",
    "active_model",
    "cell_model",
]



from typing import Dict
from typing import Optional
from typing import Tuple

import dolfin
import numpy as np
import pulse
from cardiac_geometries.geometry import MeshTypes

from . import utils
from .geometry import BaseGeometry

logger = utils.getLogger(__name__)


def create_boxmesh(lx, ly, lz, dx=0.5, refinements=0):
    # Create computational domain [0, lx] x [0, ly] x [0, lz]
    # with resolution prescribed by benchmark or more refinements

    N = lambda v: int(np.rint(v))
    mesh = dolfin.BoxMesh(
        dolfin.MPI.comm_world,
        dolfin.Point(0.0, 0.0, 0.0),
        dolfin.Point(lx, ly, lz),
        N(lx / dx),
        N(ly / dx),
        N(lz / dx),
    )

    for i in range(refinements):
        logger.info(f"Performing refinement {i + 1}")
        mesh = dolfin.refine(mesh, redistribute=False)

    return mesh


class SlabGeometry(BaseGeometry):
    @staticmethod
    def default_markers() -> Dict[str, Tuple[int, int]]:
        return {
            "X0": (2, 1),
            "X1": (2, 2),
            "Y0": (2, 3),
            "Y1": (2, 4),
            "Z0": (2, 5),
            "Z1": (2, 6),
        }

    def _default_microstructure(
        self,
        mesh: dolfin.Mesh,
        ffun: dolfin.MeshFunction,
    ) -> pulse.Microstructure:
        from cardiac_geometries import slab_fibers

        return slab_fibers.create_microstructure(
            function_space=self.parameters["fiber_space"],
            mesh=mesh,
            ffun=ffun,
            markers=self.markers,
            alpha_endo=self.parameters["fibers_angle_endo"],
            alpha_epi=self.parameters["fibers_angle_epi"],
        )

    def _default_ffun(self, mesh: dolfin.Mesh) -> dolfin.MeshFunction:
        return create_slab_facet_function(
            mesh=mesh,
            lx=self.parameters["lx"],
            ly=self.parameters["ly"],
            lz=self.parameters["lz"],
            markers=self.markers,
        )

    def _default_mesh(self) -> dolfin.Mesh:
        return create_boxmesh(
            **{
                k: v
                for k, v in self.parameters.items()
                if k in ["lx", "ly", "lz", "dx"]
            },
        )

    @staticmethod
    def default_parameters():
        return dict(
            lx=2.0,
            ly=0.7,
            lz=0.3,
            dx=0.2,
            fibers_angle_endo=0,
            fibers_angle_epi=0,
            fiber_space="Quadrature_3",
            num_refinements=1,
            mesh_type=MeshTypes.slab.value,
        )

    def validate(self):
        pass

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"lx={self.parameters['lx']}, "
            f"ly={self.parameters['ly']}, "
            f"lz={self.parameters['lz']}, "
            f"dx={self.parameters['dx']}, "
            f"num_refinements={self.parameters['num_refinements']})"
        )


def create_slab_facet_function(
    mesh: dolfin.Mesh,
    lx: float,
    ly: float,
    lz: float,
    markers: Optional[Dict[str, Tuple[int, int]]] = None,
) -> dolfin.MeshFunction:
    if markers is None:
        markers = SlabGeometry.default_markers()
    # Define domain to apply dirichlet boundary conditions
    x0 = dolfin.CompiledSubDomain("near(x[0], 0) && on_boundary")
    x1 = dolfin.CompiledSubDomain("near(x[0], lx) && on_boundary", lx=lx)
    y0 = dolfin.CompiledSubDomain("near(x[1], 0) && on_boundary")
    y1 = dolfin.CompiledSubDomain("near(x[1], ly) && on_boundary", ly=ly)
    z0 = dolfin.CompiledSubDomain("near(x[2], 0) && on_boundary")
    z1 = dolfin.CompiledSubDomain("near(x[2], lz) && on_boundary", lz=lz)

    ffun = dolfin.MeshFunction("size_t", mesh, mesh.topology().dim() - 1)
    ffun.set_all(0)

    x0.mark(ffun, markers["X0"][1])
    x1.mark(ffun, markers["X1"][1])

    y0.mark(ffun, markers["Y0"][1])
    y1.mark(ffun, markers["Y1"][1])

    z0.mark(ffun, markers["Z0"][1])
    z1.mark(ffun, markers["Z1"][1])
    return ffun


def create_slab_microstructure(fiber_space, mesh):

    family, degree = fiber_space.split("_")
    logger.debug("Set up microstructure")
    V_f = dolfin.VectorFunctionSpace(mesh, family, int(degree))
    f0 = dolfin.interpolate(
        dolfin.Constant((1, 0, 0)),
        V_f,
    )
    s0 = dolfin.interpolate(
        dolfin.Constant((0, 1, 0)),
        V_f,
    )
    n0 = dolfin.interpolate(
        dolfin.Constant((0, 0, 1)),
        V_f,
    )
    # Collect the microstructure
    return pulse.Microstructure(f0=f0, s0=s0, n0=n0)



import math

import simcardems


def test_SlabBoundaryConditions(geo):
    bcs = simcardems.boundary_conditions.create_slab_boundary_conditions(
        geo=geo,
        pre_stretch=0.1,
        traction=1.0,
        spring=0.2,
        fix_right_plane=False,
    )

    assert len(bcs.neumann) == 1
    assert math.isclose(float(bcs.neumann[0].traction), 1.0)

    assert len(bcs.dirichlet) == 1
    assert callable(bcs.dirichlet[0])

    assert len(bcs.robin) == 1
    assert math.isclose(float(bcs.robin[0].value), 0.2)



from __future__ import annotations

from pathlib import Path
from typing import Optional
from typing import Tuple
from typing import TYPE_CHECKING
from typing import Union

import cbcbeat
import dolfin
import numpy as np
from dolfin import FiniteElement  # noqa: F401
from dolfin import MixedElement  # noqa: F401
from dolfin import tetrahedron  # noqa: F401
from dolfin import VectorElement  # noqa: F401

from ... import save_load_functions as io
from ... import utils
from ...config import Config
from ...geometry import load_geometry
from ..em_model import BaseEMCoupling
from ..em_model import setup_EM_model

if TYPE_CHECKING:
    from ...datacollector import Assigners

logger = utils.getLogger(__name__)


class EMCoupling(BaseEMCoupling):
    @property
    def coupling_type(self):
        return "pureEP_ORdmm_Land"

    @property
    def vs(self) -> dolfin.Function:
        return self.ep_solver.solution_fields()[0]

    def register_ep_model(self, solver: cbcbeat.SplittingSolver):
        logger.debug("Registering EP model")
        self.ep_solver = solver
        logger.debug("Done registering EP model")

    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, type(self)):
            return NotImplemented

        if not super().__eq__(__o):
            return False

        if not np.allclose(self.vs.vector().get_local(), __o.vs.vector().get_local()):
            return False
        return True

    @property
    def assigners(self) -> Assigners:
        return self._assigners

    @assigners.setter
    def assigners(self, assigners) -> None:
        self._assigners = assigners

    def setup_assigners(self) -> None:
        from ...datacollector import Assigners

        self.assigners = Assigners(vs=self.vs, mech_state=None)
        for name, index in [
            ("V", 0),
            ("Ca", 45),
            ("XS", 40),
            ("XW", 41),
            ("CaTrpn", 42),
            ("TmB", 43),
            ("Cd", 44),
        ]:
            self.assigners.register_subfunction(
                name=name,
                group="ep",
                subspace_index=index,
            )

    def solve_ep(self, interval: Tuple[float, float]) -> None:
        self.ep_solver.step(interval)

    def print_ep_info(self):
        # Output some degrees of freedom
        total_dofs = self.vs.function_space().dim()
        logger.info("EP model")
        utils.print_mesh_info(self.ep_mesh, total_dofs)

    def cell_params(self):
        return self.ep_solver.ode_solver._model.parameters()

    @property
    def ep_mesh(self):
        return self.geometry.ep_mesh

    def update_prev_ep(self):
        self.ep_solver.vs_.assign(self.ep_solver.vs)

    def save_state(
        self,
        path: Union[str, Path],
        config: Optional[Config] = None,
    ) -> None:

        super().save_state(path=path, config=config)

        with dolfin.HDF5File(
            self.geometry.comm(),
            Path(path).as_posix(),
            "a",
        ) as h5file:
            h5file.write(self.ep_solver.vs, "/ep/vs")

        io.dict_to_h5(
            self.cell_params(),
            path,
            "ep/cell_params",
        )

    @classmethod
    def from_state(
        cls,
        path: Union[str, Path],
        drug_factors_file: Union[str, Path] = "",
        popu_factors_file: Union[str, Path] = "",
        disease_state="healthy",
        PCL: float = 1000,
    ) -> BaseEMCoupling:
        logger.debug(f"Load state from path {path}")
        path = Path(path)
        if not path.is_file():
            raise FileNotFoundError(f"File {path} does not exist")

        geo = load_geometry(path, schema_path=path.with_suffix(".json"))
        logger.debug("Open file with h5py")
        with io.h5pyfile(path) as h5file:
            config = Config(**io.h5_to_dict(h5file["config"]))
            state_params = io.h5_to_dict(h5file["state_params"])
            cell_params = io.h5_to_dict(h5file["ep"]["cell_params"])
            vs_signature = h5file["ep"]["vs"].attrs["signature"].decode()

        config.drug_factors_file = drug_factors_file
        config.popu_factors_file = popu_factors_file
        config.disease_state = disease_state
        config.PCL = PCL

        VS = dolfin.FunctionSpace(geo.ep_mesh, eval(vs_signature))
        vs = dolfin.Function(VS)
        logger.debug("Load functions")
        with dolfin.HDF5File(geo.ep_mesh.mpi_comm(), path.as_posix(), "r") as h5file:
            h5file.read(vs, "/ep/vs")

        from . import CellModel, ActiveModel

        cell_inits = io.vs_functions_to_dict(
            vs,
            state_names=CellModel.default_initial_conditions().keys(),
        )

        return setup_EM_model(
            cls_EMCoupling=cls,
            cls_CellModel=CellModel,
            cls_ActiveModel=ActiveModel,
            geometry=geo,
            config=config,
            cell_inits=cell_inits,
            cell_params=cell_params,
            state_params=state_params,
        )
